from .tuner import SSFConfig
from typing import TYPE_CHECKING, Any, Dict
from .deltatuner_model import DeltaTunerModel, DelatunerModelForCausalLM


DELTATUNER_TYPE_TO_CONFIG_MAPPING = {
    "SSF": SSFConfig,
}

MODEL_TYPE_TO_DELTATUNER_MODEL_MAPPING = {
    "CAUSAL_LM": DelatunerModelForCausalLM
}

def get_delta_config(config_dict: Dict[str, Any]):
    """
    Returns a Delta Tuner config object from a dictionary.

    Args:
        config_dict (`Dict[str, Any]`): Dictionary containing the configuration parameters.
    """

    return DELTATUNER_TYPE_TO_CONFIG_MAPPING[config_dict["peft_type"]](**config_dict)



import sys
import os
import json
import time
import timeit
import yaml
import random
import torch
from transformers import AutoTokenizer
from transformers.generation.utils import GenerationConfig


class Timer:
    level = 0

    def __init__(self, name):
        self.name = name

    def __enter__(self):
        self.start = timeit.default_timer()
        Timer.level += 1

    def __exit__(self, *a, **kw):
        Timer.level -= 1
        print(f'{"  " * Timer.level}{self.name} took {timeit.default_timer() - self.start} sec')

def network_latency(model, tokenizer, batch_size=1, max_seq_length=32, infer_cnt=3.):
    batch, _ = input_constructor(batch_size, max_seq_length, tokenizer)
    aver_time = 0.
    model.eval()
    for i in range(int(infer_cnt)):
        start = time.time()
        with torch.no_grad():
            _ = model(**batch)
        end = time.time()
        sep = 1000 * (end - start)
        if i == 0:
            continue
        else:
            aver_time += sep / (infer_cnt)
    return aver_time

def input_constructor(batch_size, resolution, tokenizer):
    #tokenizer.pad_token = tokenizer.eos_token
    prompt = ["Once upon a time",
              "It was in the",
              "In his lifetime and",
              "the close links that",
              "In 2000 Boulter had",
              "a noted politician and",
              "He had an elder",
              "which he was attracted",
    ]
    prompt = prompt[:batch_size]
    inputs = tokenizer(prompt, return_tensors="pt", truncation=True)
    return inputs, tokenizer

def network_is_legal(cand, vis_dict, params, supernet):
    if cand not in vis_dict:
        vis_dict[cand] = {}
    info = vis_dict[cand]
    if 'visited' in info:
        return False
    cand = json.loads(cand)
    num_layers = sum(cand["num_hidden_layers"])
    if num_layers < 1:
        return False
    candidate_net = supernet.set_sample_config(cand)
    n_parameters = sum(param.numel() for param in candidate_net.parameters() if param.requires_grad)
    info['params'] = n_parameters / 10.**6
    if params.max_param_limits and info['params'] > params.max_param_limits:
        return False
    if params.min_param_limits and info['params'] < params.min_param_limits:
        return False
    info['visited'] = True
    return True

def populate_random_func(search_space, denas_config, supernet, search_space_name):
    cand_tuple = dict()
    for name in search_space_name:
        cand_tuple[name] = []
    for name in search_space_name:
        for i in range(getattr(supernet.config, denas_config.layer_name)):
            cand_tuple[name].append(random.choice(search_space[f"{name}_{i}"]))
    return json.dumps(cand_tuple)

def mutation_random_func(m_prob, s_prob, search_space, supernet, denas_config, top_candidates, search_space_name):
    cand = random.choice(top_candidates)
    result_cand = json.loads(cand)
    for i in range(getattr(supernet.config, denas_config.layer_name)):
        random_s = random.random()
        if random_s < s_prob:
            result_cand["num_hidden_layers"][i] = random.choice(search_space["num_hidden_layers_{}".format(i)])
    for name in search_space_name:
        if name == "num_hidden_layers":
            continue
        for i in range(getattr(supernet.config, denas_config.layer_name)):
            random_s = random.random()
            if random_s < m_prob:
                result_cand[name][i] = random.choice(search_space[f"{name}_{i}"])
    return json.dumps(result_cand)

def crossover_random_func(top_candidates, denas_config, supernet, search_space_name):
    p1 = json.loads(random.choice(top_candidates))
    p2 = json.loads(random.choice(top_candidates))
    max_iters_tmp = 50
    while p1 != p2 and max_iters_tmp > 0:
        max_iters_tmp -= 1
        p1 = json.loads(random.choice(top_candidates))
        p2 = json.loads(random.choice(top_candidates))
    cand = {}
    for name in search_space_name:
        cand[name] = []
    for name in search_space_name:
        for i in range(getattr(supernet.config, denas_config.layer_name)):        
            cand[name].append(random.choice([p1[name][i], p2[name][i]]))
    return json.dumps(cand)


import logging
import gc
import json
import numpy as np
from transformers import AutoTokenizer

from abc import ABC, abstractmethod
from .utils import network_latency, network_is_legal, populate_random_func
from ..scores import do_compute_nas_score

 
class BaseSearchEngine(ABC):
    def __init__(self, params=None, super_net=None, search_space=None, peft_type=None):
        super().__init__()
        self.super_net = super_net
        self.search_space = search_space
        self.params = params
        self.peft_type=peft_type
        if self.params.tokenizer is None:
            raise RuntimeError("Please specify the right tokenizer in the deltatuner algo!")
        self.tokenizer =  self.params.tokenizer
        logging.basicConfig(level=logging.INFO,
                format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        self.logger = logging.getLogger('DENAS')
    
    @abstractmethod
    def search(self):
        pass

    @abstractmethod
    def get_best_structures(self):
        pass

    '''
    Judge sample structure legal or not
    '''
    def cand_islegal(self, cand):
        return network_is_legal(cand, self.vis_dict, self.params, self.super_net)

    '''
    Check whether candidate latency is within latency budget
    '''
    def cand_islegal_latency(self, cand):
        if hasattr(self.params, "budget_latency_max") or hasattr(self.params, "budget_latency_min"):
            latency = self.get_latency(cand)
            if hasattr(self.params, "budget_latency_max") and self.params.budget_latency_max is not None and self.params.budget_latency_max < latency:
                return False
            if hasattr(self.params, "budget_latency_min") and self.params.budget_latency_min is not None and self.params.budget_latency_min > latency:
                return False
        return True

    '''
    Compute candidate latency
    '''
    def get_latency(self, cand):
        if 'latency' in self.vis_dict[cand]:
            return self.vis_dict[cand]['latency']
        latency = np.inf
        cand_dict = json.loads(cand)
        #model = self.super_net.set_sample_config(cand_dict)
        latency = network_latency(self.super_net, self.tokenizer, batch_size=self.params.batch_size)
        self.vis_dict[cand]['latency']= latency
        return self.vis_dict[cand]['latency']

    '''
    Generate sample random structure
    '''
    def populate_random_func(self):
        return populate_random_func(self.search_space, self.params, self.super_net, self.params.search_space_name)

    '''
    Compute nas score for sample structure
    '''
    def cand_evaluate(self, cand):
        cand_dict = json.loads(cand)
        #model = self.super_net.set_sample_config(cand_dict)
        nas_score, score, latency = do_compute_nas_score(model=self.super_net, 
                                                        tokenizer=self.tokenizer,
                                                        resolution=self.params.img_size,
                                                        batch_size=self.params.batch_size,
                                                        mixup_gamma=1e-2,
                                                        expressivity_weight=self.params.expressivity_weight,
                                                        complexity_weight=self.params.complexity_weight,
                                                        diversity_weight=self.params.diversity_weight,
                                                        saliency_weight=self.params.saliency_weight,
                                                        latency_weight=self.params.latency_weight,
                                                        peft_type=self.peft_type)
        self.vis_dict[cand]['score'] = nas_score
        return nas_score, score, latency


from .config import DeltaTunerType, TRANSFORMERS_MODELS_TO_SSF_TARGET_MODULES_MAPPING, BEST_MODEL_STRUCTURE_DEFAULT_NAME
from .save_and_load import get_deltatuner_model_state_dict, set_deltatuner_model_state_dict



import os, sys, time, math
import torch
from torch import nn
import numpy as np
import gc
import torch
from torchsummary import summary
from torch.nn.modules import linear
from transformers import pytorch_utils

from .utils import ATTN_LAYER_STRUCTURE_NAME, LINEAR_LAYER_STRUCTURE_NAME
from ..search.utils import network_latency, input_constructor

# TODO: separate domain specific ops
def get_linear_layer_metric_array(net, metric):
    metric_array = []
    for k, param in net.named_parameters():
        if not any(name in k for name in LINEAR_LAYER_STRUCTURE_NAME):
            continue
        else:
            if "ssf_" in k:
                metric_array.append(metric(param))
    return metric_array

def get_attn_layer_metric_array(net, metric):
    metric_array = []
    for k, param in net.named_parameters():
        if not any(name in k for name in ATTN_LAYER_STRUCTURE_NAME):
            continue
        else:
            if "ssf_" in k:
                metric_array.append(metric(param))
    return metric_array

def compute_diversity_score(net):
    
    # select the gradients that we want to use for search/prune
    def nuc_norm(param):
        if param.grad is not None:
            return torch.linalg.vector_norm(param) * torch.linalg.vector_norm(param.grad)
        else:
            return torch.zeros_like(param)

    diversity_score_list = get_attn_layer_metric_array(net, nuc_norm)
    return diversity_score_list

def compute_saliency_score(net):
    # convert params to their abs. Keep sign for converting it back.
    @torch.no_grad()
    def linearize(net):
        signs = {}
        for name, param in net.state_dict().items():
            signs[name] = torch.sign(param)
            #param.abs_()
        return signs
    # convert to orig values
    @torch.no_grad()
    def nonlinearize(net, signs):
        for name, param in net.state_dict().items():
            if 'weight_mask' not in name:
                param.mul_(signs[name])

    # keep signs of all params
    #signs = linearize(net)
    # select the gradients that we want to use for search/prune
    def synflow(param):
        if param.grad is not None:
            return torch.abs(param * param.grad)
        else:
            return torch.zeros_like(param)

    grads_abs = get_linear_layer_metric_array(net, synflow)
    grads_abs.extend(get_attn_layer_metric_array(net, synflow))
    # apply signs of all params
    #nonlinearize(net, signs)
    return grads_abs



import torch
from peft import LoraModel
from peft.import_utils import is_bnb_available, is_bnb_4bit_available
from peft.tuners.lora import _get_submodules, mark_only_lora_as_trainable, _freeze_adapter, LoraLayer, Embedding, Conv2d, Linear, Conv1D

if is_bnb_available():
    import bitsandbytes as bnb
    from peft.tuners.lora import Linear8bitLt
    if is_bnb_4bit_available():
        from peft.tuners.lora import Linear4bit

class DeltaLoraSearchSpace:
    @classmethod
    def generate_search_space(cls, model, denas_config,
                              max_r=12, r_step=1, 
                              max_alpha=6, alpha_step=1):
        search_space = {}
        search_space_name = ["num_hidden_layers", "r", "alpha"]
        for i in range(getattr(model.config, denas_config.layer_name)):
            search_space["num_hidden_layers_{}".format(i)] = [0,1]
            search_space["r_{}".format(i)] = list(range(r_step, max_r+r_step, r_step))
            search_space["alpha_{}".format(i)] = list(range(alpha_step, max_alpha+alpha_step, alpha_step))
        return search_space, search_space_name
    
    @classmethod
    def generate_supernet(cls, model, denas_config, search_space):
        supernet_config = {}
        supernet_config["num_hidden_layers"] = [search_space["num_hidden_layers_{}".format(i)][-1] for i in range(getattr(model.config, denas_config.layer_name))]
        supernet_config["r"] = [search_space["r_{}".format(i)][-1] for i in range(getattr(model.config, denas_config.layer_name))]
        supernet_config["alpha"] = [search_space["alpha_{}".format(i)][-1] for i in range(getattr(model.config, denas_config.layer_name))]
        return supernet_config

class DeltaLoraModel(LoraModel):
    def __init__(self, model, config, adapter_name, model_param):
        super().__init__(model, config, adapter_name)
        self.model = model
        self.forward = self.model.forward
        self.peft_config = config
        self.model_param = model_param
        self.adapter_name = adapter_name
        self.set_sample_config(self.model_param)

    def _create_org_module(self, target):
        bias = hasattr(target, "bias") and target.bias is not None
        loaded_in_4bit = getattr(self.model, "is_loaded_in_4bit", False)
        loaded_in_8bit = getattr(self.model, "is_loaded_in_8bit", False)

        if loaded_in_8bit and isinstance(target, Linear8bitLt):
            eightbit_kwargs = {}
            eightbit_kwargs.update(
                {
                    "has_fp16_weights": target.state.has_fp16_weights,
                    "memory_efficient_backward": target.state.memory_efficient_backward,
                    "threshold": target.state.threshold,
                    "index": target.index,
                }
            )
            new_module = bnb.nn.Linear8bitLt(input_features=target.in_features, output_features=target.out_features, bias=bias, **eightbit_kwargs)
        elif loaded_in_4bit and isinstance(target, Linear4bit):
            fourbit_kwargs = {}
            fourbit_kwargs.update(
                {
                    "compute_dtype": target.compute_dtype,
                    "compress_statistics": target.weight.compress_statistics,
                    "quant_type": target.weight.quant_type,
                }
            )
            new_module = bnb.nn.Linear4bit(target.in_features, target.out_features, bias=bias, **fourbit_kwargs)
        elif isinstance(target, Embedding):
            new_module = torch.nn.Embedding(num_embeddings=target.num_embeddings, embedding_dim=target.embedding_dim)
        elif isinstance(target, Conv2d):
            out_channels, in_channels = target.weight.size()[:2]
            kernel_size = target.weight.size()[2:]
            stride = target.stride
            padding = target.padding
            new_module = torch.nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=kernel_size, stride=stride, padding=padding)
        else:
            if isinstance(target, Linear):
                in_features, out_features = target.in_features, target.out_features
            elif isinstance(target, Conv1D):
                in_features, out_features = (target.weight.ds_shape if hasattr(target.weight, "ds_shape") else target.weight.shape)
            elif isinstance(target, torch.nn.Linear):
                in_features, out_features = target.in_features, target.out_features
            else:
                raise ValueError(
                    f"Target module {target} is not supported. ")
            new_module = torch.nn.Linear(in_features, out_features, bias=bias)
        return new_module

    def _replace_module(self, parent_module, child_name, new_module, old_module):
        setattr(parent_module, child_name, new_module)
        if old_module.weight.shape != new_module.weight.shape:
            old_module.weight.data = old_module.weight.data.T
        new_module.weight = old_module.weight
        if hasattr(old_module, "bias"):
            if old_module.bias is not None:
                new_module.bias = old_module.bias

        if getattr(old_module, "state", None) is not None:
            new_module.state = old_module.state
            new_module.to(old_module.weight.device)

        # dispatch to correct device
        for name, module in new_module.named_modules():
            if "lora_" in name:
                module.to(old_module.weight.device)
            if "ranknum" in name:
                module.to(old_module.weight.device)

    def set_sample_config(self, model_config):
        lora_config = self.peft_config[self.adapter_name]
        self._check_quantization_dependency()
        is_target_modules_in_base_model = False
        key_list = [key for key, _ in self.model.named_modules()]
        
        for key in key_list:
            if not self._check_target_module_exists(lora_config, key):
                continue

            is_target_modules_in_base_model = True
            parent, target, target_name = _get_submodules(self.model, key)
            
            l_num = [int(k) for k in key.split(".") if k.isnumeric()].pop()
            r = model_config["r"][l_num]
            alpha = model_config["alpha"][l_num] * r
            if model_config["num_hidden_layers"][l_num]:
                lora_config.r = r
                lora_config.lora_alpha = alpha
                if isinstance(target, LoraLayer) and isinstance(target, torch.nn.Conv2d):
                    target.update_layer_conv2d(self.adapter_name, lora_config.r, lora_config.lora_alpha, lora_config.lora_dropout, lora_config.init_lora_weights,)
                elif isinstance(target, LoraLayer) and isinstance(target, torch.nn.Embedding):
                    target.update_layer_embedding(self.adapter_name, lora_config.r, lora_config.lora_alpha, lora_config.lora_dropout, lora_config.init_lora_weights,)
                elif isinstance(target, LoraLayer):
                    target.update_layer(self.adapter_name, lora_config.r, lora_config.lora_alpha, lora_config.lora_dropout, lora_config.init_lora_weights,)
                else:
                    new_module = self._create_new_module(lora_config, self.adapter_name, target)
                    self._replace_module(parent, target_name, new_module, target)
            else:
                new_module = self._create_org_module(target)
                self._replace_module(parent, target_name, new_module, target)

        if not is_target_modules_in_base_model:
            raise ValueError(
                f"Target modules {lora_config.target_modules} not found in the base model. "
                f"Please check the target modules and try again.")
        mark_only_lora_as_trainable(self.model, self.peft_config[self.adapter_name].bias)
        if self.peft_config[self.adapter_name].inference_mode:
            _freeze_adapter(self.model, self.adapter_name)
        return self.model



from typing import List
from dataclasses import dataclass, field
from transformers import AutoTokenizer

@dataclass
class DeltaTunerArguments:
    search_engine: str = field(default="EvolutionarySearchEngine", metadata={"help": "Which search strategy used in the DE-NAS. Can be 'EvolutionarySearchEngine', 'RandomSearchEngine' and 'SigoptSearchEngine'"})
    model_id: str = field(default=None, metadata={"help": "The model id in the hugging face or the local model path"})
    batch_size: int = field(default=4, metadata={"help": "The batch size used in the search process to calculate the inference"})
    max_epochs: int = field(default=10, metadata={"help": "The max epochs of the search processing"})
    scale_factor: int = field(default=10, metadata={"help": "The scaling factor used in the EA search"})
    select_num: int = field(default=50, metadata={"help": "The num of top candidates stored in the EA search process"})
    population_num: int = field(default=50, metadata={"help": "The num of population candidates used in the EA/Random search process"})
    m_prob: float = field(default=0.2, metadata={"help": "The m probability to produce new mutation network depth"})
    s_prob: float = field(default=0.4, metadata={"help": "The s probability to produce new other mutation network structures"})
    crossover_num: int = field(default=25, metadata={"help": "The number of crossover in the EA search process"})
    mutation_num: int = field(default=25, metadata={"help": "The number of mutation in the EA search process"})
    img_size: int = field(default=32, metadata={"help": "The input or output sequence length of the model."})
    expressivity_weight: float = field(default=0, metadata={"help": "The weight of expressivity score in the de-score"})
    complexity_weight: float = field(default=0, metadata={"help": "The weight of complexity score in the de-score"})
    diversity_weight: float = field(default=1, metadata={"help": "The weight of diversity score in the de-score"})
    saliency_weight: float = field(default=1, metadata={"help":"The weight of saliency score in the de-score"})
    latency_weight: float = field(default=0, metadata={"help":"The weight of latency score in the de-score"})
    max_param_limits: int = field(default=None, metadata={"help": "The max M number of trainable parameter size"})
    min_param_limits: int = field(default=None, metadata={"help": "The min M number of trainable parameter size"})
    budget_latency_max: int = field(default=None, metadata={"help": "The max latency limit of the candidate network (etc., '3*10**6' second for your estimation of max latency of LLM forward)"})
    budget_latency_min: int = field(default=None, metadata={"help": "The min latency limit of the candidate network (etc., '1*10**6' second for your estimation of min latency of LLM forward)"})
    layer_name: str = field(default="num_hidden_layers", metadata={"help": "The layer name for a specific model search space"})
    best_model_structure: str = field(default=None, metadata={"help": "The file path to load the best model structure"})
    search_space_name: List[str] = field(default=None, metadata={"help": "The search space name of delta layers"})
    random_seed: int = field(default=12345, metadata={"help": "The random seed of the deltatuner search process"})
    denas: bool = field(default=True, metadata={"help": "Whether to use the denas"})
    ssf_target_module: List[str] = field(default_factory=lambda: None, metadata={"help": "Target modules for the SSF method."},)
    task_type: str = field(default="CAUSAL_LM", metadata={"help": "The task type of the deltatuner model"})
    algo: str = field(default="auto", metadata={"help": "The base algorithm of the deltatuner model, current support algorithms including: lora, ssf, auto"})


import os
import sys
import traceback
from .BaseSearchEngine import BaseSearchEngine
from .utils import mutation_random_func, crossover_random_func

class EvolutionarySearchEngine(BaseSearchEngine):

    def __init__(self, params=None, super_net=None, search_space=None,peft_type=None):
        super().__init__(params,super_net,search_space,peft_type)
        self.candidates = []
        self.top_candidates = []
        self.vis_dict = {}

    '''
    Higher order function that stack random candidates
    '''
    def stack_random_cand(self, random_func, batchsize=10):
        while True:
            cands = [random_func() for _ in range(batchsize)]
            for cand in cands:
                if cand not in self.vis_dict:
                    self.vis_dict[cand] = {}
            for cand in cands:
                yield cand

    '''
    EA mutation function for random structure
    '''
    def mutation_random_func(self):
        return mutation_random_func(self.params.m_prob, self.params.s_prob, self.search_space, self.super_net, self.params, self.top_candidates, self.params.search_space_name)

    '''
    EA crossover function for random structure
    '''
    def crossover_random_func(self):
        return crossover_random_func(self.top_candidates, self.params, self.super_net, self.params.search_space_name)

    '''
    Supernet decoupled EA populate process
    '''
    def get_populate(self):
        res = []
        max_iters = self.params.scale_factor * self.params.population_num 
        cand_iter = self.stack_random_cand(self.populate_random_func)
        while len(res) < self.params.population_num and max_iters > 0:
            max_iters -= 1
            cand = next(cand_iter)
            if not self.cand_islegal(cand):
                continue
            if not self.cand_islegal_latency(cand):
                continue
            self.cand_evaluate(cand)
            res.append(cand)
            self.logger.info('random {}/{} structure {} nas_score {} params {}'.format(len(res), self.params.population_num, cand, self.vis_dict[cand]['score'], self.vis_dict[cand]['params']))
        self.logger.info('random_num = {}'.format(len(res)))
        self.candidates += res

    '''
    Supernet decoupled EA mutation process
    '''
    def get_mutation(self):
        res = []
        max_iters = self.params.scale_factor * self.params.mutation_num  
        cand_iter = self.stack_random_cand(self.mutation_random_func)
        while len(res) < self.params.mutation_num and max_iters > 0:
            max_iters -= 1
            cand = next(cand_iter)
            if not self.cand_islegal(cand):
                continue
            if not self.cand_islegal_latency(cand):
                continue
            self.cand_evaluate(cand)
            res.append(cand)
            self.logger.info('mutation {}/{} structure {} nas_score {} params {}'.format(len(res), self.params.mutation_num, cand, self.vis_dict[cand]['score'], self.vis_dict[cand]['params']))
        self.logger.info('mutation_num = {}'.format(len(res)))
        return res

    '''
    Supernet decoupled EA crossover process
    '''
    def get_crossover(self):
        res = []
        max_iters = self.params.scale_factor * self.params.crossover_num
        cand_iter = self.stack_random_cand(self.crossover_random_func)
        while len(res) < self.params.crossover_num and max_iters > 0:
            max_iters -= 1
            cand = next(cand_iter)
            if not self.cand_islegal(cand):
                continue
            if not self.cand_islegal_latency(cand):
                continue
            self.cand_evaluate(cand)
            res.append(cand)
            self.logger.info('crossover {}/{} structure {} nas_score {} params {}'.format(len(res), self.params.crossover_num, cand, self.vis_dict[cand]['score'], self.vis_dict[cand]['params']))
        self.logger.info('crossover_num = {}'.format(len(res)))
        return res

    '''
    Keep top candidates of select_num
    '''
    def update_population_pool(self):
        t = self.top_candidates
        t += self.candidates
        t.sort(key=lambda x: self.vis_dict[x]['score'], reverse=True)
        self.top_candidates = t[:self.params.select_num]

    '''
    Unified API for EvolutionarySearchEngine
    '''
    def search(self):
        for epoch in range(self.params.max_epochs):
            self.logger.info('epoch = {}'.format(epoch))
            self.get_populate()
            self.update_population_pool()
            mutation = self.get_mutation()
            crossover = self.get_crossover()
            self.candidates = mutation + crossover
            
        self.update_population_pool()
        return str(self.top_candidates[0])

    '''
    Unified API to get best searched structure
    '''
    def get_best_structures(self):
        best_structure = self.top_candidates[0]
        self.logger.info('best structure {} nas_score {} params {}'.format(best_structure, self.vis_dict[best_structure]['score'], self.vis_dict[best_structure]['params']))
        return best_structure